/*  load.c

   5 December 2001
   modify from s1 sim 1997, 1998

   add print, printc  to output buffer  31 Dec 2001
   modify to read mspec					14 Feb 2005
   for Sx cpu simulator					11 July 2006
*/

#include "sx.h"

#define NIL 0

#define 	PRIVATE		static

int clock, ninst;
FILE *fp;

int CP, DP;		// end of CS, end of loaded DS
int tok;
extern char token[];

symt symtab[200];
int nsym;

void error(char *s){
	printf("error: %s\n",s);
	exit(0);
}

// dump n words starting from ads
void dump(int ads, int n){
  int i;
  for(i=ads; i<ads+n; i++)
    printf("%4d %d\n",i,M[i]);
}

PRIVATE void report(void){
  float cpi;
  if( ninst == 0 )
	  printf("0 instruction\n");
  else {
  	  cpi = ((float) clock)/ninst;
  	  printf("%d instructions, %d clocks, CPI %4.2f\n",
	  	ninst,clock,cpi);
  }
}

// static alloc memory of size a
int xalloc(int a){
	int b;
	b = DP;
	if( DP + a >= MAXDS ){
		printf("%d ",DP+a);
		error("xalloc: out of memory");
	}
	DP += a;
	return b;
}

//extern char *cp2;

PRIVATE void popts(void){
	TS = M[SP];
	SP--;
}

PRIVATE void pushts(int x){
	SP++;
	M[SP] = TS;
	TS = x;
}

// copy C string to M[tok]
PRIVATE void copyToken(void){
	int i;
	i = 0;
	while(token[i] != 0){
		M[tok+i] = (int)token[i];
		i++;
	}
	M[tok+i] = 0;
}

// ------ print non-lin machine code --------

PRIVATE int isAtom(int e){ return (e & 0x80000000) != 0;}
PRIVATE int head(int e){return M[e];}
PRIVATE int tail(int e){return M[e+1];}

PRIVATE void pA1(char *s){ printf("%s ",s); }
PRIVATE void pA2(char *s, int v){ printf("%s.%d ",s,v); }

PRIVATE void prAtom( int e){
	int op,arg;
	op = (e >> 24) & 0x7f;
	arg = e & 0x0ffffff;
	switch(op){
	case  xIF: pA1("if"); break;
	case  xWHILE: pA1("while"); break;
	case  xDO: pA1("do"); break;
	case  xNEW: pA1("new"); break;
	case  xADD: pA1("+"); break;
	case  xSUB: pA1("-"); break;
	case  xMUL: pA1("*"); break;
	case  xDIV: pA1("/"); break;
	case  xBAND: pA1("&"); break;
	case  xSHR: pA1(">>"); break;
	case  xSHL: pA1("<<"); break;
	case  xEQ: pA1("="); break;
	case  xLT: pA1("<"); break;
	case  xGT: pA1(">"); break;
	case  xCALL: pA2("call",arg); break;
	case  xGET: pA2("get",arg); break;
	case  xPUT: pA2("put",arg); break;
	case  xLIT: pA2("lit",arg); break;
	case  xLDX: pA2("ldx",arg); break;
	case  xSTX: pA2("stx",arg); break;
	case  xLD: pA2("ld",arg); break;
	case  xST: pA2("st",arg); break;
	case  xLDY: pA2("ldy",arg); break;
	case  xSTY: pA2("sty",arg); break;
	case  xFUN:	printf("fun.%d.%d ",arg>>8,arg&0x0ff); break;
	case  xSYS: pA2("sys",arg); break;
	case  xSTR: pA2("str",arg); break;
	}
}

PRIVATE void prL2(int e){
	if( e == NIL ) return;
	if( isAtom(e) )
		prAtom(e);
	else{
		prList(head(e));
		prL2(tail(e));
	}
}

void prList( int e){
	if( e == NIL ) return;
	if( isAtom(e) )
		prAtom(e);
	else{
		printf("(");
		prL2(e);
		printf(")");
	}
}

void trap(int op, int arg){		/* special function */
	int n;
	switch(op) {
	case icEnd :
		report();
		runflag = 0;
		break;
	case icArray:
		TS = xalloc(TS);
		break;
	case icSys:
		switch(arg){
		case 1: 		// print ts
// 	  		n = sprintf(cp2,"%d", TS);
			printf("%d",TS);
			popts();
//	  		cp2 += n;
	  		break;
		case 2 : 		// printc ts
//			n = sprintf(cp2,"%c", TS);
			printf("%c",TS);
			popts();
//			cp2 += n;
	  		break;
		case 3:
			tokenise();
			copyToken();
			pushts(tok);
			break;
		case 9: pushts(DP); break;
		case 10:
			prList(TS);
			popts();
			break;
		case 11: readinfile(); break;
//		case 12: printf(" line %d ",line); break;
		case 13:
			runflag = 0;
			break;
		default:
			printf("%d ",arg);
			error("unknown syscall");
		}
		break;
	default:
		error("unknown trap");
	}
}

void show(void){		/* show some registers */
  int i;
  printf("%d:%d ",PC,mpc);
//  disassem();
  printf("TS %d FP %d SP %d NX %d FF %d IR %d tbus %d dbus %d Z%d S%d\n",
    TS,FP,SP,NX,FF,IR,tbus,dbus,Z,S);
}

// read microprogram mpgm.txt
void readmpgm(void){
  int i,n;
  FILE *fp;
  char ibuf[1000];

  fp = fopen("mpgm.txt","r");
  fgets(ibuf,1000,fp);
  sscanf(ibuf,"%d",&n);	  /* length of mpgm */
  for( i=0; i<n; i++ ) {
    fgets(ibuf,1000,fp);
    strncpy(ms,ibuf+4,mwidth);  // be careful ibuf+4
	scan1(i);
  }
  fclose(fp);
}

// read mspec.h to get microaddress
void readmspec(void){
	int i,n,v;
	FILE *fp;
	char w1[80],w2[80];

	fp = fopen("mspec.h","r");
	i = 0;
	while(1){
		fscanf(fp,"%s%s%d",w1,w2,&v);
		if( eqs(w2,"MLEN") ){	// last symbol
			mlen = v;
			break;
		}else if( eqs(w2,"MAWIDTH") ){
			mawidth = v;
		}else if( eqs(w2,"MCWIDTH") ){
			mcwidth = v;
		}else if( w2[0] == 'a' ){
			strcpy(mads[i].name,w2+2);
			mads[i].val = v;
			i++;
		}
	}
	mads[i].name[0] = 0;			// end of symbol
	mwidth = mawidth + mcwidth;
//	for(n=0; n<i; n++)
//		printf("%s %d\n",mads[n].name,mads[n].val);
//	printf("ma %d mc %d mlen %d\n",mawidth,mcwidth,mlen);
}

PRIVATE int readM(int *mem, FILE *fi){
	int i, b, e, c;
	fscanf(fi,"%d%d",&b,&e);
	for(i=b; i<=e; i++){
		fscanf(fi,"%d",&c);
		mem[i] = c;
	}
	return e;
}
/*
// change arg at M[ads]
PRIVATE void patch(int ads, int arg){
	M[ads] = (arg << 8) | (M[ads] & 255);
}

// relocate CS from a..b, offset by MAXDS
PRIVATE void reloc(int start, int end){
	int i, op, arg;
	for(i=start; i<=end; i++){
		op = M[i] & 255;
		arg = M[i] >> 8;
		switch(op){
		case icLd:
		case icSt:
		case icCall: patch(i,arg+MAXDS); break;
		}
	}
}
*/
void readsymtab(FILE *fi){
	int i, ty, ref, d;
	char nm[80];
	fscanf(fi,"%d",&nsym);
	for(i=1; i<=nsym; i++){
		fscanf(fi,"%s%d%d%d%d",nm,&ty,&ref,&d,&d);
		strcpy(symtab[i].name,nm);
		symtab[i].type = ty;
		symtab[i].ref = ref;
	}
}

void dumpsym(void){
	int i;
	printf("<%d> ",nsym);
	for(i=1; i<=nsym; i++)
		printf("%s %d %d\n",
		symtab[i].name,symtab[i].type,symtab[i].ref);
}

char *findName(int ty, int ref){
	int i;
	for(i=1; i<=nsym; i++)
		if(symtab[i].type==ty && symtab[i].ref==ref)
			break;
	return (i>nsym) ? 0 : symtab[i].name;
}

// read obj from stdin into memory
void loadobj(char *fn){
	int a;
	FILE *fi;
	fi = fopen(fn,"r");
	if(fi == NULL){
	 printf("input file not found");
	 exit(0);
	}
	fscanf(fi,"%d",&a);
	if( a != OBJ_SOMv2 )
		error("not a som-v2 object");
	a = readM(M,fi);	// code
	CP = a + 1;
	a = readM(M,fi);	 	// string constant, data
	DP = a + 1;
	printf("load program, last address %d\n",CP);
	printf("DP %d\n",DP);
	readsymtab(fi);
	fclose(fi);
//	reloc(MAXDS,CP-1);
}
